;-------------------------------------------------------------
; Creates nested directories. For example if hPath is
; "C:\nested\path" it will create both \nested and \path
; directories.
; returns TRUE on success and FALSE on failure
;-------------------------------------------------------------
proc CreateNestedDir uses ebx ecx edx esi edi, hPath
        stdcall StrPtr, [hPath]
        or      eax,eax
        jz      .finish
        mov     esi,eax
        mov     edi,eax
        stdcall StrLen,eax
        or      eax,eax
        jz      .finish
.scan:
        mov     al,[esi]
        cmp     al,'\'
        je      .dir
        cmp     al,'/'
        je      .dir
        or      al,al
        jz      .dir
        inc     esi
        jmp     .scan
.dir:                   ; test if directory exists, create new one if not
        mov     bl,al
        mov     byte[esi],0
        stdcall DirExists, edi
        or      eax,eax
        jnz     .direxists
        ; attempt to create directory
        invoke  CreateDirectory, edi,0
        mov     [esi],bl
        or      eax,eax
        jz      .finish
.direxists:
        or      bl,bl
        jz      .return_true
        mov     byte[esi],bl
        inc     esi
        jmp     .scan
.return_true:
        xor     eax,eax
        dec     eax
.finish:
        ret
endp


;---------------------------------------------------------------------
; Converts some filename's path (full or relative) in relative
; form, related to current directory.
; Arg:
;   hFileName - handle/ptr of the filename.
; Returns:
;   eax - Handle to new string containing the relative path/filename
;---------------------------------------------------------------------
proc GetRelativePath .hFileName
locals
.root rb 512
.name rb 512
.res  rb 512
.ptr  dd ?
endl

        push    esi edi ebx

        lea      esi, [.root]
        invoke  GetCurrentDirectory, 512, esi
        cmp     byte [esi+eax-1], '\'
        je      @f
        mov     word [esi+eax], '\'
@@:
        stdcall StrPtr, [.hFileName]
        lea     edi, [.name]
        lea     ecx, [.ptr]
        invoke  GetFullPathName, eax, 512, edi, ecx
        test    eax, eax
        jz      .returnempty    ; there is invalid or missing filename.

        lea     edx, [.res]

; find first subdir
        mov     cl, ':'
        call    _CmpSubFolder
        jc      .copy                   ; Copy full path, because they are in different drives.

        mov     cl, '\'

        inc     esi
        inc     edi

.subloop:
        call    _CmpSubFolder
        jnc     .subloop

        dec     esi

; count of remaining subfolders.
.count:
        inc     esi
        mov     al, [esi]
        test    al,al
        je      .copy
        cmp     al,cl
        jne     .count
.found:
        mov     dword [edx], '..\'
        add     edx, 3
        test    al,al
        jnz     .count

.copy:
        mov     al, [edi]
        inc     edi
        mov     [edx], al
        inc     edx
        test    al,al
        jnz     .copy

        stdcall StrNew
        lea     edx, [.res]
        stdcall StrCopy, eax, edx

        pop     ebx edi esi
        ret

.returnempty:
        stdcall StrNew
        pop     ebx edi esi
        ret
endp




proc _CmpSubFolder       ; esi - first, edi - second, cl - end char

        xor     ebx, ebx

.scanloop:
        mov     al, [esi+ebx]
        test    al,al
        jz      .endofpath

        mov     ch, al
        mov     ah, [edi+ebx]
        or      ax, $2020
        cmp     al, ah
        jne     .notequal

        cmp     ch,cl
        je      .eosubdir

        inc     ebx
        jmp     .scanloop

.endofpath:
        cmp     byte [edi+ebx], cl
        jne     .notequal
        inc     edi
        add     edi, ebx
        add     esi, ebx
        xor     ebx, ebx
        stc
        ret

.eosubdir:
        add     esi, ebx
        add     edi, ebx
        inc     esi
        inc     edi
        xor     ebx, ebx
        clc
        ret

.notequal:
        stc
        ret
endp


;----------------------------------------
; Returns a pointer to the filename
; separated from full path name.
; The original string remains unchanged
;----------------------------------------
proc ExtractFileName hPath
        push    ecx edx
        stdcall StrPtr, [hPath]
        or      eax,eax
        jz      .end
        mov     edx, eax
        xor     ecx, ecx
.loop:
        mov     cl, [edx]
        inc     edx
        jecxz   .end
        cmp     cl, '\'
        je      .dir
        cmp     cl, '/'
        jne     .loop
.dir:
        mov     eax, edx
        jmp     .loop
.end:
        pop     edx ecx
        ret
endp

;-----------------------------------------
; Returns a pointer to the file extension
; separated from full path and file name.
; The original string remains unchanged
;-----------------------------------------
proc ExtractExt uses ecx edx esi, hPath
        stdcall StrPtr, [hPath]
        or      eax,eax
        jz      .finish
        mov     edx,eax
        mov     esi,eax
        stdcall StrLen, eax
        or      eax,eax
        jz      .finish
        add     edx,eax
        mov     ecx,edx
        dec     edx
.scan:
        mov     al,[edx]
        cmp     al,'.'
        je      .found
        cmp     al,'\'
        je      .notfound
        cmp     al,'/'
        je      .notfound
        cmp     esi,edx
        je      .notfound
        dec     edx
        jmp     .scan
.notfound:
        mov     eax,ecx
        jmp     .finish
.found:
        mov     eax,edx
.finish:
        ret
endp

;----------------------------------------------------
; Strips file name from given path.
;----------------------------------------------------
proc ExtractDirectory uses esi ecx, hPath, hBuffer
local new_str_flag:DWORD
        mov     [new_str_flag],0
        mov     eax,[hBuffer]
        cmp     eax,[hPath]
        je      .buffer_ok
        or      eax,eax                 ; if hBuffer is NULL, then
        jnz     .buffer_not_null        ; create new string
        stdcall StrNew
        or      eax,eax
        jz      .finish
        mov     [new_str_flag],0
        mov     [hBuffer],eax
  .buffer_not_null:
        stdcall StrCopy, [hBuffer],[hPath]
  .buffer_ok:
        stdcall StrPtr, [hBuffer]
        mov     esi,eax
        stdcall StrLen, eax
        or      eax,eax
        jz      .finish
        mov     ecx,eax
        add     esi,ecx
        dec     esi
  .scan:
        mov     al,byte[esi]
        cmp     al,'/'
        je      .found_slash
        cmp     al,'\'
        je      .found_slash
        dec     ecx
        dec     esi
        jnz     .scan
        cmp     [new_str_flag],0
        je      @f
        stdcall StrDel, [hBuffer]
    @@: xor     eax,eax
        ret
  .found_slash:
        mov     byte[esi+1],0
        mov     eax,TRUE
  .finish:
        ret
endp

;--------------------------------------------------------------------
; Allocates needed memory and loads the entire file in this memory.
; Returns pointer to memory block or NULL if there is error.
; Returns size in ecx
;--------------------------------------------------------------------
proc LoadBinaryFile .hFileName
local .bytes:DWORD
        push    esi edi ebx
        stdcall StrPtr, [.hFileName]
        invoke  CreateFile, eax, GENERIC_READ,FILE_SHARE_READ, 0, \
                OPEN_EXISTING, 0, 0
        cmp     eax, INVALID_HANDLE_VALUE
        je      .error
        mov     edi, eax
        invoke  GetFileSize, edi, NULL
        mov     ebx, eax
        lea     ecx, [ebx+2]    ; It is for the case that the file is
                                ; text and we want to use it as null terminated
                                ; string.
        invoke  HeapAlloc, [hHeap], HEAP_ZERO_MEMORY, ecx
        mov     esi, eax
        test    esi, esi
        jz      .close
        lea     ecx, [.bytes]
        invoke  ReadFile, edi, esi, ebx, ecx, NULL
.close:
        invoke  CloseHandle, edi
        test    esi, esi
        jz      .error
        cmp     [.bytes], ebx
        je      .ok
        invoke  HeapFree, [hHeap], 0, esi
.error:
        xor     eax, eax
        xor     ecx, ecx
        pop     ebx edi esi
        ret
.ok:
        mov     eax, esi
        mov     ecx, ebx
        pop     ebx edi esi
        ret
endp



;--------------------------------------------------------------------
; Saves to the file [size] bytes from given [ptr] address.
; returns TRUE if error and FALSE if OK.
;--------------------------------------------------------------------
proc SaveBinaryFile .hFileName, .aptr, .size
local .bytes:DWORD
        push    esi edi ebx
        stdcall StrPtr, [.hFileName]
        invoke  CreateFile, eax, GENERIC_WRITE, 0, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0
        cmp     eax, INVALID_HANDLE_VALUE
        je      .error

        mov     edi, eax

        lea     ecx, [.bytes]
        invoke  WriteFile, edi, [.aptr], [.size], ecx, NULL
        test    eax, eax
        jnz     .close

        or      [.size], -1

.close:
        invoke  CloseHandle, edi

        mov     eax, [.size]
        cmp     [.bytes], eax
        je      .ok

.error:
        xor     eax, eax
        inc     eax
        pop     ebx edi esi
        ret

.ok:
        xor     eax, eax
        pop     ebx edi esi
        ret
endp





;----------------------------------------------------------
; Create an unique for the project filename of shape:
; <hPrefix>NNN<hSuffix> - where NNN is the mimimal number
; that makes the filename unique.
; Returns:
;   eax - handle to the new created filename.
;   edx - the NNN number.
;----------------------------------------------------------

proc CreateUniqueName .hPrefix, .hSuffix
locals
.str rb 1024
.num dd ?
.nm  dd ?
.ptr dd ?
endl
        push    esi edi ebx
        mov     [.num], 1
        stdcall StrNew
        mov     [.nm], eax
        lea     ebx, [.str]

; Create name
.findloop:
        stdcall StrCopy, [.nm], [.hPrefix]

        stdcall NumToStr, [.num], ebx, 0, ntsUnsigned or ntsDec or ntsZeroTerminated
        stdcall StrCat, [.nm], ebx
        stdcall StrCat, [.nm], [.hSuffix]

        stdcall StrPtr, [.nm]
        lea     esi, [.ptr]
        invoke  GetFullPathName, eax, 1024, ebx, esi

        stdcall IsFileOpen, ebx
        test    eax, eax
        jz      .found

.nextloop:
        inc     [.num]
        jmp     .findloop

.found:
        stdcall StrCopy, [.nm], ebx
        mov     eax, [.nm]
        mov     edx, [.num]

        pop     ebx edi esi
        ret
endp



proc HashFilename .hFileName
locals
.str rb 1024
.ptr dd ?
endl
        push    esi edi ebx
        lea     esi, [.str]

        stdcall StrPtr, [.hFileName]
        lea     ecx, [.ptr]
        invoke  GetFullPathName, eax, 1024, esi, ecx
        stdcall StrHash, esi, $20
        pop     ebx edi esi
        ret
endp



;-------------------------------------------
; check if given file is present on disk
; returns TRUE is so, FALSE otherwise
;-------------------------------------------
proc FileExists uses ecx, .hFileName
local .hfile:DWORD
        stdcall StrPtr, [.hFileName]
        xor     ecx, ecx
        invoke  CreateFile, eax, 0, FILE_SHARE_READ, ecx, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, ecx
        cmp     eax, INVALID_HANDLE_VALUE
        je      .notexists
        invoke  CloseHandle, eax
        xor     eax, eax
        inc     eax
        ret
.notexists:
        xor     eax, eax
        ret
endp


;-------------------------------------------
; same as FileExists, checks directories
;-------------------------------------------
proc DirExists hPath
local temp_dir[MAX_PATH]:BYTE
        lea     eax,[temp_dir]
        invoke  GetCurrentDirectory, MAX_PATH,eax
        stdcall StrPtr, [hPath]
        invoke  SetCurrentDirectory, eax
        or      eax,eax
        jz      .not_found
        lea     eax,[temp_dir]
        invoke  SetCurrentDirectory,eax
.not_found:
        ret
endp

proc ChangeDir .hFileName
locals
.ptrname dd ?
.str rb 1024
endl
        push    esi
        stdcall StrPtr, [.hFileName]

        lea     ecx, [.ptrname]
        lea     esi, [.str]
        invoke  GetFullPathName, eax, 1024, esi, ecx
        or      eax,eax
        jz      .finish
        mov     eax, [.ptrname]
        test    eax, eax
        jz      @f
        mov     byte [eax], 0
@@:
        invoke  SetCurrentDirectory, esi
        pop     esi
.finish:
        ret
endp